
/******************************************************************************
* MODULE     : compound.cc
* DESCRIPTION: Evaluation of compound expressions
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <evaluator.hh>
#include <xtable.hh>
#include <function.hh>
#include <converter.hh>
namespace mmx {

GC_GLOBAL_STATIC (converter, cvt, converter ());
GC_GLOBAL_STATIC (table, cvt_cache, table ());
GC_GLOBAL_STATIC (bool, amb_flag, false);

/******************************************************************************
* Implicit conversions for function application
* ----------------------------------------------------------------------------
* The conversion uses a rudimentary caching mechanism.
* However, this mechanism never frees cached values and
* should therefore be improved in the future.
* ----------------------------------------------------------------------------
* The routine 'apply_prog' returns a program which should be evaluated
* in order to perform the required conversions. The returned program
* necessitates additional auxiliary variables which are stored in
* symbols 0#, 1#, ... In particular, 0# contains the function and
* #1 the argument(s). The auxiliary variables are stored in the current
* environment. We might have to create a local sub-environment
* for each function application in the future (it seems that we do so now).
* New auxiliary variables can be introduced in conv_lambda.cc.
******************************************************************************/

static bool
is_really_ambiguous (generic r) {
  if ((!is_list (r)) || (N(r) != 2)) return false;
  if (!is_func (r[0], CONV_AMBIGUOUS)) return false;
  if (!is_func (r[1], ID_AND)) return false;
  table t;
  list l= cdr (as_list (r[1]));
  while (!nil (l)) {
    if (t->contains (car (l))) return true;
    t (car (l))= generic ();
    l= cdr (l);
  }
  return false;
}

static void
make_disambiguous (list& progs, list& types, table& t) {
  while (!nil (types) && t->contains (car (types))) {
    progs= cdr (progs);
    types= cdr (types);
  }
  if (nil (types)) return;
  t (car (types))= generic ();
  make_disambiguous (cdr (progs), cdr (types), t);
}

generic
evaluator_rep::apply_prog (generic fun_type, generic arg_type) {
  generic pair= list (fun_type, arg_type);
  if (!cvt_cache->contains (pair)) {
    // cout << fun_type << " on " << arg_type << lf;
    generic r= cvt->apply_prog (fun_type, arg_type);
    // cout << "---> " << r << lf;
    if (is_really_ambiguous (r)) {
#ifdef AMBIGUOUS_FIX
      // be less regarding about ambiguous expressions
      // fixes problem with l' for l: Alias List T in combination
      // with assignment fix in match.cc
      table t;
      list progs= cddr (as_list (r[0]));
      list types= cdr (as_list (r[1]));
      make_disambiguous (progs, types, t);
      if (N(progs) == 1) r= list (car (progs), car (types));
      else r= list (cons (CONV_AMBIGUOUS, cons (ID_AND, types), progs),
		    cons (ID_AND, types));
#else
      amb_flag= true;
      error (car (trace), "Ambiguous function application");
      cerr << "Debug] apply program: " << r << lf;
      return generic (); // list (r[0][2], r[1][1]);
#endif
    }
    // cout << "---> " << r << lf;
    cvt_cache (pair)= r;
  }
  return cvt_cache [pair];
}

/******************************************************************************
* Function application
******************************************************************************/

void
evaluator_rep::kw_compound_1 () {
  // Input: (compound_1, val_first, ...)
  generic next= car (rem);
  if (gentype (next) == KEYWORD_TYPE) {
    cur= next;
    rem= cdr (rem);
    // Output: (keyw, ...)
    (this ->* (as_keyword (next)->fn)) ();
  }
  else {
    list orig= as_list (car (trace));
    cur= KW_COMPOUND_2;
    rem= cons (next, list (), cdr (orig), cdr (rem));
    // Output: (compound_2, first_arg, compound_2, (func), rem_args, ...)
    kw_compound_2 ();
  }
}

void
evaluator_rep::kw_compound_2 () {
  // Input: (compound_2, val_next, rev_vals_prev, rem_args, cont, ...)
  list vals_done= cons (car (rem), as_list (cadr (rem)));
  list rem_args = as_list (caddr (rem));
  if (nil (rem_args)) {
    vals_done   = reverse (vals_done);
    value   fun = as_value (car (vals_done));
    list    args= cdr (vals_done);
    generic arg = cons (MMX_TUPLE, get_values (args));
    generic type= cons (ID_CROSS, get_types (args));
    generic conv= apply_prog (fun->type, type);
    // cout << "conv= " << conv << "\n";
    if (nil (conv)) {
      if (!amb_flag) {
	cout << "fun= " << fun << "\n";
	cout << "args= " << args << "\n";
	error (car (trace), "Invalid function application");
      }
      amb_flag= false;
      return;
    }
    table xtra= table ();
    xtra (CONV_FUNCTION)= fun;
    xtra (CONV_ARGUMENT)= value (arg, type);
    start_local_env (xtra);
    cur= conv[0];
    rem= cons (KW_COMPOUND_3, cdddr (rem));
    // Output: (conv_prog, kw_compound_3, cont, ...)
  }
  else {
    cur= car (rem_args);
    rem= cons (KW_COMPOUND_2, vals_done, cdr (rem_args), cdddr (rem));
    // Output: (next_arg, compound_2, rev_vals_prev, rem_args, cont, ...)
  }
}

void
evaluator_rep::kw_compound_3 () {
  // Input: (compound_3, result, cont, ...)
  end_local_env ();
  cur= cadr (rem);
  rem= cons (car (rem), cddr (rem));
  // Output: (cont, result, ...)
  trace= cdr (trace);
}

} // namespace mmx
