
/******************************************************************************
* MODULE     : ext_parse.cc
* DESCRIPTION: Parse extension code
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extend.hh>
#include <literal.hh>
#include <single.hh>
#include <file.hh>
#include <error.hh>
#include <parser.hh>
namespace mmx {

/******************************************************************************
* Rewrite before parsing
******************************************************************************/

generic
extender_rep::rewrite (generic g) {
  // ungroup "import", "export" and "forall" statements
  // normalize function types
  // normalize literals
  if (is_func (g, MMX_BEGIN)) {
    list l= reverse (cdr (as_list (g))), r;
    while (!nil (l)) {
      generic h= rewrite (car (l));
      if (is_func (h, MMX_BEGIN)) r= cdr (as_list (h)) * r;
      else r= cons (h, r);
      l= cdr (l);
    }
    return cons (MMX_BEGIN, r);
  }
  else if (is_func (g, MMX_IMPORT, 1) ||
	   is_func (g, MMX_EXPORT, 1) ||
	   is_func (g, MMX_FORALL, 2))
    {
      generic h= g[N(g)-1];
      if (is_func (h, MMX_LOCAL) && is_func (h[N(h)-1], MMX_BEGIN))
	h= cons (MMX_BEGIN,
		 cDr (cdr (as_list (h))) * cdr (as_list (h[N(h)-1])));
      h= rewrite (h);
      list head= cDr (as_list (g));
      if (is_func (h, MMX_BEGIN)) {
	list l= reverse (cdr (as_list (h))), r;
	while (!nil (l)) {
	  r= cons (head * list (car (l)), r);
	  l= cdr (l);
	}
	return cons (MMX_BEGIN, r);
      }
      return head * list (generic (h));
    }
  else if (is_func (g, MMX_FUNCTION, 2)) {
    generic dom= rewrite (g[1]);
    generic im = rewrite (g[2]);
    if (dom == MMX_VOID) dom= list (MMX_CROSS);
    if (im == MMX_VOID) im= list (MMX_CROSS);
    if (is_func (dom, MMX_TUPLE)) dom= cons (MMX_CROSS, cdr (as_list (dom)));
    if (is_func (im , MMX_TUPLE)) im = cons (MMX_CROSS, cdr (as_list (im )));
    if (!is_func (dom, MMX_CROSS)) dom= list (MMX_CROSS, dom);
    return list (car (g), dom, im);
  }
  else if (is_list (g)) {
    list l= reverse (as_list (g)), r;
    while (!nil (l)) {
      r= cons (rewrite (car (l)), r);
      l= cdr (l);
    }
    return r;
  }
  else if (is_literal (g)) {
    generic h= S (as_literal (g));
    if (is_symbol (h) && is_string (S (as_symbol (h)))) {
      string s= as_string (S (as_symbol (h)));
      return symbol (unquote (s));
    }
    else return h;
  }
  else return g;
}

/******************************************************************************
* Type declarations
******************************************************************************/

static bool
is_type_type (generic g) {
  if (is_func (g, MMX_FORALL, 2)) return is_type_type (g[2]);
  if (is_func (g, MMX_FUNCTION, 2)) return is_type_type (g[2]);
  return g == MMX_CLASS;
}

static bool
is_type_decl (generic g) {
  if (is_func (g, MMX_FORALL, 2)) return is_type_decl (g[2]);
  if (is_func (g, MMX_DEFINE, 2)) return is_type_decl (g[1]);
  if (is_func (g, MMX_DEFINEALIAS, 2)) return is_type_decl (g[1]);
  if (is_func (g, MMX_ASSIGN, 2)) return is_type_decl (g[1]);
  if (is_func (g, MMX_TYPE, 2)) return is_type_type (g[2]);
  return false;
}

static int
type_args_sub (generic g) {
  if (is_func (g, MMX_FORALL, 2)) return type_args_sub (g[2]);
  if (is_func (g, MMX_FUNCTION, 2)) {
    int count= 0;
    list l (g[1]);
    if (is_func (g[1], MMX_CROSS)) l= cdr (as_list (g[1]));
    while (!nil (l)) {
      generic h= car (l);
      if (h == MMX_CLASS) count++;
      else if (is_func (h, MMX_TYPE, 2) && (h[2] == MMX_CLASS)) count++;
      l= cdr (l);
    }
    return count;
  }
  return 0;
}

static int
type_args (generic g) {
  if (is_func (g, MMX_FORALL, 2)) return type_args (g[2]);
  if (is_func (g, MMX_DEFINE, 2)) return type_args (g[1]);
  if (is_func (g, MMX_DEFINEALIAS, 2)) return type_args (g[1]);
  if (is_func (g, MMX_ASSIGN, 2)) return type_args (g[1]);
  if (is_func (g, MMX_TYPE, 2)) return type_args_sub (g[2]);
  return 0;
}

void
extender_rep::ex_type_decl (generic g) {
  // Exported type declarations should be of one of the following form
  //   T: Class == "cpp_name"
  //   T: (X_1, ..., X_n) -> Class == "cpp_name"
  // Each parameter X_i is either an ordinary type or equal to Class

  if (!is_func (g, MMX_DEFINE, 2))
    error (g, "Expected a declaration");
  else if (!is_func (g[1], MMX_TYPE, 2))
    error (g, "Expected a declaration");
  else if (!is_symbol (g[1][1]))
    error (g[1][1], "Expected a symbol");
  else if (is_func (g[1][2], MMX_FUNCTION, 2) &&
	   (!is_func (g[1][2][1], MMX_CROSS)))
    error (g[1][2][1], "Expected a cross");
  else {
    max_type_args= max (max_type_args, type_args (g));
    if (max_type_args > MAX_TYPE_ARGS)
      error (g, "Too many type arguments");
    generic var= get_decl_var (g);
    generic val= get_decl_val (g);
    if (ex_type_conv -> contains (var)) {
      if (ex_type_conv[var] != val)
	error (g, "Re-exported type as a different symbol");
    }
    else {
      ex_types= cons (g, ex_types);
      ex_type_conv (var)= val;
      if (top_level) ex_type_top (g)= single (true);
    }
  }
}

/******************************************************************************
* Analyzing function declarations
******************************************************************************/

generic
get_decl_var (generic g) {
  if (is_func (g, MMX_FORALL, 2)) return get_decl_var (g[2]);
  if (is_func (g, MMX_DEFINE, 2)) return get_decl_var (g[1]);
  if (is_func (g, MMX_DEFINEALIAS, 2)) return get_decl_var (g[1]);
  if (is_func (g, MMX_ASSIGN, 2)) return get_decl_var (g[1]);
  if (is_func (g, MMX_TYPE, 2)) return g[1];
  return generic ();
}

generic
get_decl_type (generic g) {
  if (is_func (g, MMX_FORALL, 2))
    return list (g[0], g[1], get_decl_type (g[2]));
  if (is_func (g, MMX_DEFINE, 2)) return get_decl_type (g[1]);
  if (is_func (g, MMX_DEFINEALIAS, 2)) return get_decl_type (g[1]);
  if (is_func (g, MMX_ASSIGN, 2)) return get_decl_type (g[1]);
  if (is_func (g, MMX_TYPE, 2)) return g[2];
  return generic ();
}

generic
get_decl_val (generic g) {
  if (is_func (g, MMX_FORALL, 2)) return get_decl_val (g[2]);
  if (is_func (g, MMX_DEFINE, 2)) return g[2];
  if (is_func (g, MMX_DEFINEALIAS, 2)) return g[2];
  if (is_func (g, MMX_ASSIGN, 2)) return g[2];
  return generic ();
}

/******************************************************************************
* Function declarations
******************************************************************************/

static bool
inconsistent_redefinition (generic what, list l) {
  generic val= get_value (what, l);
  if (!nil (val)) return get_decl_val (what) != val;
  return false;
}

void
extender_rep::gen_import_decl (generic g) {
  // Generic importations should all be of the form
  //   forall(T:Class) f: (A_1,...,A_n)->B == "f";
  // Genericity over more than one type is not yet implemented.
  // If one among A_1,...,A_n,B is of the form F(T),
  // where F(T) involves T, but is not equal to T or Alias T,
  // then a converter F(X)->F(Y) should be supplied for all X and Y
  // for which a converter X->Y exists.

  if (!is_func (g, MMX_FORALL, 2))
    error (g, "Expected a 'forall' type");
  else if (!is_func (g[1], MMX_TYPE, 2))
    error (g[1], "Expected a typed quantification variable");
  else if (g[1][2] != MMX_CLASS)
    error (g[1][2], "Expected 'Class' type");
  else if (is_func (g[2], MMX_FORALL, 2))
    error (g, "Double quantifications not yet allowed");
  else if (!is_func (g[2], MMX_DEFINE, 2))
    error (g[2], "Expected a declaration");
  else if (!is_func (g[2][1], MMX_TYPE, 2))
    error (g[2], "Expected a declaration");
  else {
    generic var = get_decl_var  (g);
    generic type= get_decl_type (g);
    if (nil (var) || nil (type)) return;
    if (match_type (type, gen_import_env [var])) {
      if (inconsistent_redefinition (g, gen_imports))
	error (g, "Re-imported generic routine as a different symbol");
      return;
    }
    if (gen_import_env -> contains (var)) {
      generic t= gen_import_env [var];
      if (!is_func (t, MMX_AND)) t= list (MMX_AND, t);
      t= cons (MMX_AND, type, cdr (as_list (t)));
      gen_import_env (var)= t;
    }
    else gen_import_env (var)= type;

    gen_imports= cons (g, gen_imports);

    if (gen_import_start) {
      if (gen_import_top != top_level)
	error (g, "Generic routines cannot be both in "
	          "included and top-level files");
    }
    else {
      gen_import_start= true;
      gen_import_top  = top_level;
    }
  }
}

void
extender_rep::export_decl (generic g) {
  // Exportations should be routines or templates of the form
  //   forall(T_1:U_n)...forall(T_n:U_n) f: (A_1,...,A_n)->B == "f";
  // Constants cannot be exported yet

  generic h= g;
  while (is_func (h, MMX_FORALL, 2)) h= h[2];
  if (!is_func (h, MMX_DEFINE, 2))
    error (h, "Expected a declaration");
  else if (!is_func (h[1], MMX_TYPE, 2))
    error (h, "Expected a declaration");
  else if (!is_symbol (h[2]))
    error (h[2], "Expected a symbol");
  else if (!is_func (h[1][2], MMX_FUNCTION, 2))
    error (h, "Constant exports not yet implemented");
  else if (!is_func (h[1][2][1], MMX_CROSS))
    error (h[1][2][1], "Cross expected");
  else {
    generic var = get_decl_var  (g);
    generic type= get_decl_type (g);
    if (nil (var) || nil (type)) return;
    if (match_type (type, export_env [var])) {
      if (inconsistent_redefinition (g, exports))
	error (g, "Re-exported routine as a different symbol");
      return;
    }
    if (export_env -> contains (var)) {
      generic t= export_env [var];
      if (!is_func (t, MMX_AND)) t= list (MMX_AND, t);
      t= cons (MMX_AND, type, cdr (as_list (t)));
      export_env (var)= t;
    }
    else export_env (var)= type;
    exports= cons (g, exports);
    if (top_level) export_top (g)= single (true);
  }
}

/******************************************************************************
* Parsing (construct types, exports, imports)
******************************************************************************/

generic
extender_rep::load (string name) {
  if (cur_file != "" && !starts (name, "/")) {
    for (int i= N(cur_file)-1; i>=0; i--)
      if (cur_file[i] == '/') {
	name= cur_file (0, i+1) * name;
	break;
      }
  }
  cur_file= name;
  if (loaded->contains (name)) {
    // cout << "  Duplicate " << name << lf;
    return Begin ();
  }
  else loaded (name)= single (true);
  // cout << "  Processing " << name << lf;
  if (top_level) {
    generic sname= list (mmx_symbol ("string"), get_file_name (name));
    top_files = top_files * list (sname);
  }
  else {
    generic call_init= mmx_symbol ("call_initializer");
    generate_init (list (call_init, the_prefix, get_file_name (name)));
  }

  string s;
  insert_source (name, "cmdline");
  if (mmx::load (name, s)) {
    cerr << "Couldn't open file " << name << lf;
    exit (1);
  }
  list errors;
  generic prog= mmx::parse (s, errors);
  if (!nil (errors)) {
    print_errors (errors);
    exit (1);
  }
  if (!is_func (prog, MMX_BEGIN))
    prog= Begin (prog);
  return prog;
}

void
extender_rep::parse (generic g, bool reverse_flag) {
  g= rewrite (g);
  if (is_func (g, MMX_BEGIN)) {
    list l= cdr (as_list (g));
    while (!nil (l)) {
      generic h= car (l);
      if (is_func (h, MMX_IMPORT, 1)) {
	/*if (is_type_decl (h[1])) import_type_decl (h[1]);
	else*/ gen_import_decl (h[1]);
      }
      if (is_func (h, MMX_EXPORT, 1)) {
	if (is_type_decl (h[1])) ex_type_decl (h[1]);
	else export_decl (h[1]);
      }
      if (is_func (h, symbol ("preamble"), 1))
	preamble= preamble * as_string (S(h[1]));
      if (is_func (h, symbol ("postamble"), 1))
	postamble= postamble * as_string (S(h[1]));
      if (is_func (h, symbol ("load"), 1)) {
	string name= as_string (S(h[1]));
	string old_cur_file= cur_file;
	parse (load (name), false);
	cur_file= old_cur_file;
      }
      if (is_func (h, symbol ("include"), 1)) {
	string name= as_string (S(h[1]));
	string old_cur_file= cur_file;
	bool old_top_level= top_level;
	top_level= false;
	parse (load (name), false);
	top_level= old_top_level;
	cur_file= old_cur_file;
      }
      l= cdr (l);
    }
  }

  if (reverse_flag) {
    gen_imports= reverse (gen_imports);
    ex_types= reverse (ex_types);
    exports= reverse (exports);
    im_types= reverse (im_types);
    imports= reverse (imports);
    //nice_print ("Generic imports", gen_imports);
    //nice_print ("Exported types", ex_types);
    //nice_print ("Exported routines", exports);
    //nice_print ("Imported types", im_types);
    //nice_print ("Imported routines", imports);
    //cout << "gen_import_env= " << gen_import_env << lf;
    //cout << "ex_type_conv  = " << ex_type_conv << lf;
    //cout << "export_env    = " << export_env << lf;
    //cout << "im_type_conv  = " << im_type_conv << lf;
    //cout << "import_env    = " << import_env << lf;
    //cout << "max_type_args = " << max_type_args << lf;
  }
}

/******************************************************************************
* Printing the dependencies
******************************************************************************/

void
extender_rep::print_dependencies (generic g) {
  g= rewrite (g);
  if (is_func (g, MMX_BEGIN)) {
    list l= cdr (as_list (g));
    while (!nil (l)) {
      generic h= car (l);
      if (is_func (h, symbol ("load"), 1) ||
	  is_func (h, symbol ("include"), 1))
	{
	  string name= as_string (S(h[1]));
	  string old_cur_file= cur_file;
	  top_level= false;
	  generic prog= load (name);
	  cout << " " << cur_file;
	  print_dependencies (prog);
	  cur_file= old_cur_file;
	}
      l= cdr (l);
    }
  }
}

/******************************************************************************
* Adding default routines
******************************************************************************/

static generic
copy_function (generic T) {
  generic fun= Function (Cross (T), T);
  return Define (Type (symbol ("new"), fun), symbol ("new"));
}

static generic
delete_function (generic T) {
  generic fun= Function (Cross (T), MMX_VOID);
  return Define (Type (symbol ("delete"), fun), symbol ("delete"));
}

static generic
assign_function (generic T) {
  generic fun= Function (Cross (T, T), MMX_VOID);
  return Define (Type (symbol ("assign"), fun), symbol ("assign"));
}

static generic
hash_function (generic T) {
  generic fun= Function (Cross (T), MMX_INT);
  return Define (Type (symbol ("hash"), fun), symbol ("hash"));
}

static generic
print_function (generic T) {
  generic fun= Function (Cross (MMX_OSTREAM, T), MMX_OSTREAM);
  return Define (Type (symbol ("print"), fun), symbol ("operator <<"));
}

static generic
equal_function (generic T) {
  generic fun= Function (Cross (T, T), MMX_BOOL);
  return Define (Type (MMX_EQUAL, fun), symbol ("operator =="));
}

static generic
unequal_function (generic T) {
  generic fun= Function (Cross (T, T), MMX_BOOL);
  return Define (Type (MMX_UNEQUAL, fun), symbol ("operator !="));
}

static list
args_list (int n, int i=1) {
  if (i>n) return list ();
  return cons (symbol ("X" * as_string (i)), args_list (n, i+1));
}

static generic
make_template (generic body, list types, int i=1) {
  if (nil (types)) return body;
  body= make_template (body, cdr (types), i+1);
  generic X= symbol ("X" * as_string (i));
  return Forall (Type (X, car (types)), body);
}

list
extender_rep::add_default_routines (generic g, mmx_function fn) {
  if (!(is_func (g, MMX_DEFINE, 2) &&
	is_func (g[1], MMX_TYPE, 2) &&
	is_symbol (g[1][1])))
    error (g, "Assertion failed in add_default_routines");
  else {
    generic T= g[1][1];
    if (is_func (g[1][2], MMX_FUNCTION, 2))
      T= cons (T, args_list (N(g[1][2][1]) - 1));
    generic pr= fn (T);
    if (is_func (g[1][2], MMX_FUNCTION, 2))
      pr= make_template (pr, cdr (as_list (g[1][2][1])));
    return list (pr);
  }
  return list ();
}

void
extender_rep::add_default_routines () {
  generic T  = symbol ("T");
  generic TC = Type (T, MMX_CLASS);
  gen_imports= cons (Forall (TC, unequal_function (T)), gen_imports);
  gen_imports= cons (Forall (TC, equal_function (T)), gen_imports);
  gen_imports= cons (Forall (TC, print_function (T)), gen_imports);
  gen_imports= cons (Forall (TC, hash_function (T)), gen_imports);
  gen_imports= cons (Forall (TC, assign_function (T)), gen_imports);
  gen_imports= cons (Forall (TC, delete_function (T)), gen_imports);
  gen_imports= cons (Forall (TC, copy_function (T)), gen_imports);

  list l= ex_types, r;
  while (!nil (l)) {
    if (ex_type_top->contains (car (l))) {
      r= add_default_routines (car (l), copy_function) * r;
      r= add_default_routines (car (l), assign_function) * r;
      r= add_default_routines (car (l), delete_function) * r;
      r= add_default_routines (car (l), hash_function) * r;
      r= add_default_routines (car (l), print_function) * r;
      r= add_default_routines (car (l), equal_function) * r;
      r= add_default_routines (car (l), unequal_function) * r;
    }
    l= cdr (l);
  }
  while (!nil (r)) {
    export_decl (car (r));
    r= cdr (r);
  }
}

} // namespace mmx
