
/******************************************************************************
* MODULE     : extend.cc
* DESCRIPTION: Automatically generate C++ extension code for generic objects
*              This should later be used in combination with mmx-glue
*              in order to produce glue code for templates
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extend.hh>
#include <ccpp.hh>
#include <pretty.hh>
#include <error.hh>
namespace mmx {

/******************************************************************************
* Required routines for generic structures
******************************************************************************/

size_t
extender_rep::Size () {
  return sizeof (extender_rep);
}

void
extender_rep::Mark () {
  mark (the_prefix);
  mark (cur_file);
  mark (top_level);
  mark (loaded);
  mark (top_files);
  mark (glue_flag);
  mark (nr_errors);
  mark (preamble);
  mark (postamble);
  mark (gen_imports);
  mark (gen_import_start);
  mark (gen_import_top);
  mark (gen_import_env);
  mark (ex_types);
  mark (ex_type_top);
  mark (ex_type_conv);
  mark (exports);
  mark (export_top);
  mark (export_env);
  mark (im_types);
  mark (im_type_top);
  mark (im_type_conv);
  mark (imports);
  mark (import_top);
  mark (import_env);
  mark (output);
  mark (glue_init);
  mark (max_type_args);
  mark (nr_ext_tvar);
}

int
extender_rep::Type () {
  return EXTENDER_TYPE;
}

void
extender_rep::Print (ostream& out) {
  out << "#extender<>";
}

/******************************************************************************
* Error handling and code generation
******************************************************************************/

void
extender_rep::error (generic where, string message) {
  cerr << std_error (message, where);
  nr_errors++;
}

void
extender_rep::warning (generic where, string message) {
  cerr << std_warning (message, where);
}

void
extender_rep::generate (generic g) {
  output= cons (g, output);
}

void
extender_rep::generate_init (generic g) {
  glue_init= cons (g, glue_init);
}

void
extender_rep::declare (generic var, generic val, generic type) {
  generic dc  = mmx_symbol ("declare");
  generic decl= list (dc, var, val, type);
  generate_init (decl);
}

generic
extender_rep::mmx_make_list (list l) {
  if (N(l) < 5) return cons (mmx_symbol ("list"), l);
  return list (mmx_symbol ("cons"), car (l), mmx_make_list (cdr (l)));
}

/******************************************************************************
* Useful subroutines
******************************************************************************/

void
nice_print (generic text, list l) {
  cout << text << stroke << lf << indent;
  while (!nil (l)) {
    pretty_print (cout, car (l));
    cout << stroke << lf;
    l= cdr (l);
  }
  cout << unindent << "  " << lf;
}

void
nice_print (symbol text, list l) {
  nice_print ((generic) text, l);
}

generic
make_struct (generic name, list args, list inherit) {
  generic cl= MMX_CLASS;
  if (!nil (inherit)) cl= cons (CCPP_INHERIT, inherit);
  return Define (Type (name, cl), cons (CCPP_STRUCT, args));
}

generic
make_blackbox_struct (generic name, list extra) {
  generic rep  = symbol ("rep");
  generic rep2 = symbol ("rep2");
  generic vp   = Pointer (MMX_VOID);
  generic trep = Type (rep, vp);
  generic trep2= Type (rep2, vp);
  generic r1= Define (Type (name, Inline (Constructor (trep2))),
		      Initialize (generic (list (rep, rep2)), Begin ()));
  return make_struct (name, list (trep, r1) * extra);
}

generic
ccpp_type_protect (generic type) {
  if (!is_list (type)) return type;
  return cons (CCPP_SPECIALIZE, as_list (type));
}

generic
Global (generic decl) {
  if (is_func (decl, MMX_TYPE, 2) && is_func (decl[2], CCPP_STATIC, 1))
    return Global (Assign (decl, list (decl[2][1])));
  if (is_func (decl, MMX_TYPE, 2))
    return Global (Assign (decl, list (decl[2])));
  assert (is_func (decl, MMX_ASSIGN, 2));
  assert (is_func (decl[1], MMX_TYPE, 2));
  generic glob= symbol ("GC_GLOBAL");
  generic var = decl[1][1];
  generic type= decl[1][2];
  generic init= decl[2];
  if (is_func (type, CCPP_STATIC, 1)) {
    glob= symbol ("GC_GLOBAL_STATIC");
    type= type[1];
  }
  return list (glob, type, var, init);
}

/******************************************************************************
* Naming parameters in function types
******************************************************************************/

symbol
parameter (int i) {
  return string ("arg_") * as_string (i);
}

static list
name_parameters (list l, int i) {
  if (nil (l)) return l;
  else return cons (Type (parameter (i), car (l)),
		    name_parameters (cdr (l), i+1));
}

generic
name_parameters (generic g) {
  // name parameters in function type
  if (is_func (g, MMX_TYPE, 2))
    return list (g[0], g[1], name_parameters (g[2]));
  else if (is_func (g, MMX_FUNCTION, 2))
    return list (g[0], name_parameters (g[1]), g[2]);
  else if (is_func (g, MMX_CROSS))
    return cons (g[0], name_parameters (cdr (as_list (g)), 1));
  else return g;
}

/******************************************************************************
* Names
******************************************************************************/

static generic
unmacro (generic g) {
  if (starts (as_string (S (as_symbol (g))), "macro ")) {
    string s= as_string (S (as_symbol (g)));
    return symbol (s (6, N(s)));
  }
  else return g;
}

symbol
extender_rep::gentype () {
  return the_prefix * "_type";
}

symbol
extender_rep::gentype_rep () {
  return the_prefix * "_type_rep";
}

generic
extender_rep::contype (generic name) {
  return symbol (the_prefix * "_") * as_symbol (unmacro (name));
}

generic
extender_rep::contype_rep (generic name) {
  return
    symbol (the_prefix * "_") * as_symbol (unmacro (name)) * symbol ("_rep");
}

generic
extender_rep::genmethod (generic name) {
  return symbol ("gen_") * ccpp_protect (name);
}

generic
extender_rep::geninst () {
  return symbol (the_prefix * "_inst");
}

generic
extender_rep::geninst (int nr) {
  return symbol (the_prefix * "_inst_") * symbol (as_string (nr));
}

generic
extender_rep::curtype (int nr) {
  return symbol (the_prefix * "_cur_") * symbol (as_string (nr));
}

/******************************************************************************
* Further names
******************************************************************************/

generic
extender_rep::mmx_scope (generic g) {
  return Scope (symbol ("mmx"), g);
}

generic
extender_rep::mmx_symbol (symbol s) {
  return mmx_scope (s);
}

generic
extender_rep::mmx_specialize (symbol s, generic a1) {
  return mmx_scope (Specialize (s, a1));
}

generic
extender_rep::mmx_specialize (symbol s, generic a1, generic a2) {
  return mmx_scope (Specialize (s, a1, a2));
}

generic
extender_rep::mmx_specialize (symbol s, generic a1, generic a2, generic a3) {
  return mmx_scope (Specialize (s, a1, a2, a3));
}

generic
extender_rep::mmx_generic () {
  return mmx_symbol ("generic");
}

generic
extender_rep::mmx_mirror () {
  return mmx_symbol ("mirror");
}

symbol
extender_rep::ext_cv () {
  return symbol (the_prefix * "_cv");
}

symbol
extender_rep::ext_id (int i) {
  return string ("ID_") * as_string (S (unmacro (ex_types[i-1][2])));
  // return string ("ID_") * as_string (i);
}

symbol
extender_rep::ext_fun (int i) {
  return string ("FUN_") * as_string (i);
}

symbol
extender_rep::ext_conv (int i) {
  return string ("CONV_") * as_string (i);
}

symbol
extender_rep::ext_tvar (int i) {
  if (i > nr_ext_tvar) nr_ext_tvar= i;
  return string ("TMPL_") * as_string (i);
}

symbol
extender_rep::ext_arg (int i) {
  return string ("x_") * as_string (i);
}

/******************************************************************************
* Interface
******************************************************************************/

void
print_dependencies (string ofile, string prefix, generic prog) {
  extender ext (prefix);
  cout << ofile << ":";
  ext->print_dependencies (prog);
  cout << lf;
}

void
extender_rep::produce (ostream& out, generic prog) {
  out << "\n/*******************************************************\n";
  out << "* This file was automatically generated by mmx-extend\n";
  out << "*******************************************************/\n\n";
  if (glue_flag) out << "#include \"glued.hh\"\n\n";
  parse (prog);
  if (glue_flag) produce_glue_type (ex_types, 1, true);
  if (!nil (gen_imports)) add_default_routines ();
  if (!nil (gen_imports)) {
    int i;
    produce_geninst ();
    produce_gentype ();
    for (i=1; i<=MAX_TYPE_ARGS; i++) {
      if (gen_import_top) generate (Global (Type (curtype (i), gentype ())));
      else generate (Type (curtype (i), Extern (gentype ())));
    }
    for (i=1; i<=max_type_args; i++)
      produce_geninst (i);
    produce_contype (ex_types, 1);
  }
  out << preamble << lf;
  out << lf;
  ccpp_print (out, cons (MMX_BEGIN, reverse (output)));
  if (glue_flag) {
    out << lf;
    output= list ();
    produce_glue ();
    ccpp_print (out, cons (MMX_BEGIN, reverse (output)));
  }
  out << lf << postamble;
}

bool
extension_code (ostream& out, string prefix, generic prog, bool glue_flag) {
  extender ext (prefix);
  ext->glue_flag    = glue_flag;
  ext->produce (out, prog);
  return ext->nr_errors > 0;
}

} // namespace mmx
