
/******************************************************************************
* MODULE     : convert.cc
* DESCRIPTION: proving type conversions
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <converter.hh>
#include <single.hh>
namespace mmx {

/******************************************************************************
* Selection of most specific methods in cases of ambiguity
******************************************************************************/

bool
converter_rep::freely_converts (generic from, generic to, int mode)
  // checks whether 'from' can be converted to 'to' for any substitution
  // of the free variables occurring in 'from' and 'to'
{
  proof p= convert (from, to, mode);
  return (!nil (p)) && (N(p->inst) == 0);
}

list
converter_rep::disambiguate (list l, int mode) {
  if (nil (l) || nil (cdr (l))) return l;
  generic p= car (l)[0];
  list d= disambiguate (cdr (l), mode), r;
  while (!nil (d)) {
    generic q= car (d)[0];
    // cout << "Comparing " << p << ", " << q << lf << indent;
    bool pq, qp;
    if (is_implies (p) && is_implies (q) && (p[1] == q[1])) {
      pq= freely_converts (p[2], q[2], mode);
      qp= freely_converts (q[2], p[2], mode);
    }
    else if (is_implies (p) && is_implies (q) && (p[2] == q[2])) {
      qp= freely_converts (p[1], q[1], mode);
      pq= freely_converts (q[1], p[1], mode);
    }
    else pq= qp= false;
    // cout << unindent << "pq=" << pq << ", qp=" << qp << lf;
    if (pq && (!qp));
    else if (qp && (!pq)) return d * r;
    else r= cons (car (d), r);
    d= cdr (d);
  }
  return cons (car (l), r);
}

list
converter_rep::disambiguate_hack (list l, int mode) {
  // FIXME: this routine is only used by converter_rep::apply_prog and
  //        this kind of disambiguities should really be treated otherwise
  if (nil (l) || nil (cdr (l))) return l;
  proof p= as_proof (car (l));
  list d= disambiguate_hack (cdr (l), mode), r;
  while (!nil (d)) {
    proof q= as_proof (car (d));
    // cout << "Comparing " << p->f << ", " << q->f << lf << indent;
    bool pq= freely_converts (p->f, q->f, mode);
    bool qp= freely_converts (q->f, p->f, mode);
    // cout << unindent << "pq=" << pq << ", qp=" << qp << lf;
    if (pq && (!qp));
    else if (qp && (!pq)) return d * r;
    else r= cons (car (d), r);
    d= cdr (d);
  }
  return cons (car (l), r);
}

static list
merge_solutions (list l) {
  if (nil (l)) return list ();
  return as_list (car(l)[1]) * merge_solutions (cdr (l));
}

/******************************************************************************
* Conversion of boolean type constructors
******************************************************************************/

proof
converter_rep::convert_from_and (generic from, generic to, int mode) {
  int  i= 1;
  list l= cdr (as_list (from)), r;
  while (!nil (l)) {
    proof p= convert (car (l), to, mode);
    if (!nil (p)) {
      list rule (symbol ("and=>"), single (i));
      proof add (from, to, rule, list (p));
      list deps= is_ambiguous (add)? add->deps: list (add);
      r= cons (list (implies (car (l), to), deps), r);
    }
    i++;
    l= cdr (l);
  }
  if ((mode & CV_COLLECT) == 0) r= disambiguate (r, mode);
  list sols= merge_solutions (r);
  return ambiguous_proof (implies (from, to), sols);
}

proof
converter_rep::convert_to_or (generic from, generic to, int mode) {
  int  i= 1;
  list l= cdr (as_list (to)), r;
  while (!nil (l)) {
    proof p= convert (from, car (l), mode);
    if (!nil (p)) {
      list rule (symbol ("=>or"), single (i));
      proof add (from, to, rule, list (p));
      list deps= is_ambiguous (add)? add->deps: list (add);
      r= cons (list (implies (car (l), to), deps), r);
    }
    i++;
    l= cdr (l);
  }
  if ((mode & CV_COLLECT) == 0) r= disambiguate (r, mode);
  list sols= merge_solutions (r);
  return ambiguous_proof (implies (from, to), sols);
}

proof
converter_rep::convert_from_or (generic from, generic to, int mode) {
  list l= cdr (as_list (from)), deps;
  while (!nil (l)) {
    proof p= convert (car (l), to, mode);
    if (nil (p)) return proof ();
    deps= deps * list (p);
    l= cdr (l);
  }
  return proof (from, to, "or=>", deps);
}

proof
converter_rep::convert_to_and (generic from, generic to, int mode) {
  list l= cdr (as_list (to)), deps;
  while (!nil (l)) {
    proof p= convert (from, car (l), mode);
    if (nil (p)) return proof ();
    deps= deps * list (p);
    l= cdr (l);
  }
  return proof (from, to, "=>and", deps);
}

/******************************************************************************
* Handling of ambiguities in the case of quantified type constructors
******************************************************************************/

static list
instantiate_sub (list l, generic fv) {
  if (nil (l)) return l;
  proof   p  = as_proof (car (l));
  generic val= p->inst->contains (fv)? p->inst [fv]: fv;
  p= substitute (p, fv, val);
  list pair (val, p);
  return cons (pair, instantiate_sub (cdr (l), fv));
}

static list
instantiate (proof p, generic fv) {
  if (is_ambiguous (p)) return instantiate_sub (p->deps, fv);
  else return instantiate_sub (list (p), fv);
}

static list
qf_proof_sub (generic f, string rule_s, list l) {
  if (nil (l)) return l;
  generic pair (car (l));
  generic rule= list (symbol (rule_s), pair[0]);
  proof q= as_proof (pair[1]);
  proof p (f, rule, list (q), q->inst);
  return cons (p, qf_proof_sub (f, rule_s, cdr (l)));
}

static proof
qf_proof (generic from, generic to, string rule, list sols) {
  generic f= implies (from, to);
  list l= qf_proof_sub (f, rule, sols);
  return ambiguous_proof (f, l);
}

/******************************************************************************
* Conversion of quantified type constructors
******************************************************************************/

proof
converter_rep::convert_from_exists (generic from, generic to, int mode) {
  generic var = from[1];
  generic wv  = witness_var (var);
  generic body= substitute (from[2], var, wv);
  proof p= convert (body, to, mode);
  if (nil (p)) return p;
  return proof (from, to, list (symbol ("exists=>"), wv), list (p));
}

proof
converter_rep::convert_to_forall (generic from, generic to, int mode) {
  generic var = to[1];
  generic wv  = witness_var (var);
  generic body= substitute (to[2], var, wv);
  proof p= convert (from, body, mode);
  if (nil (p)) return p;
  return proof (from, to, list (symbol ("=>forall"), wv), list (p));
}

proof
converter_rep::convert_from_forall (generic from, generic to, int mode) {
  generic var = from[1];
  generic fv  = free_var (var);
  generic body= substitute (from[2], var, fv);
  proof p= convert (body, to, mode);
  if (nil (p)) return p;
  list sols= instantiate (p, fv);
  return qf_proof (from, to, "forall=>", sols);
}

proof
converter_rep::convert_to_exists (generic from, generic to, int mode) {
  generic var = to[1];
  generic fv  = free_var (var);
  generic body= substitute (to[2], var, fv);
  proof p= convert (from, body, mode);
  if (nil (p)) return p;
  list sols= instantiate (p, fv);
  return qf_proof (from, to, "=>exists", sols);
}

/******************************************************************************
* Conversion of primitive type constructors
******************************************************************************/

proof
converter_rep::convert_function (generic from, generic to, int mode) {
  proof inverse= convert (to[1], from[1], mode);
  if (nil (inverse)) return proof ();
  proof direct = convert (from[2], to[2], mode);
  if (nil (direct)) return proof ();
  return proof (from, to, "=>arrow", list (inverse, direct));
}

proof
converter_rep::convert_cross (generic from, generic to, int mode) {
  (void) mode;
  list l1= cdr (as_list (from)), l2= cdr (as_list (to)), deps;
  while (!nil (l1)) {
    proof p= convert (car (l1), car (l2), CV_NORMAL);
    if (nil (p)) return proof ();
    deps= deps * list (p);
    l1= cdr (l1);
    l2= cdr (l2);
  }
  return proof (from, to, "=>cross", deps);
}

proof
converter_rep::convert_tuple (generic from, generic to, int mode) {
  (void) mode;
  list l1= cdr (as_list (from)), deps;
  generic repeat= cadr (as_list (to));
  while (!nil (l1)) {
    proof p= convert (car (l1), repeat, CV_NORMAL);
    if (nil (p)) return proof ();
    deps= deps * list (p);
    l1= cdr (l1);
  }
  return proof (from, to, "=>tuple", deps);
}

proof
converter_rep::convert_ignore (generic from, generic to, int mode) {
  (void) mode;
  return proof (from, to, "ignore=>", list ());
}

proof
converter_rep::convert_from_singlet (generic from, generic to, int mode) {
  proof p= convert (from[1], to, mode);
  if (nil (p)) return proof ();
  return proof (from, to, "singlet=>", list (p));
}

proof
converter_rep::convert_to_singlet (generic from, generic to, int mode) {
  proof p= convert (from, to[1], mode);
  if (nil (p)) return proof ();
  return proof (from, to, "=>singlet", list (p));
}

proof
converter_rep::convert_from_alias (generic from, generic to, int mode) {
  proof p= convert (from[1], to, mode);
  if (nil (p)) return proof ();
  return proof (from, to, "alias=>", list (p));
}

proof
converter_rep::convert_aliases (generic from, generic to, int mode) {
  (void) mode;
  list l1= cdr (as_list (from)), l2= cdr (as_list (to[1])), deps;
  while (!nil (l1)) {
    if (!is_aliastype (car (l1))) return proof ();
    proof p= convert (car (l1) [1], car (l2), CV_NORMAL);
    if (nil (p)) return proof ();
    deps= deps * list (p);
    l1= cdr (l1);
    l2= cdr (l2);
  }
  return proof (from, to, "=>aliases", deps);
}

/******************************************************************************
* Applying a function to an argument
******************************************************************************/

static list
apply_sub (list l, string s, proof on) {
  if (nil (l)) return l;
  generic pair (car (l));
  proof q= as_proof (pair[1]);
  proof p (pair[0], symbol (s), list (q, on), q->inst);
  return cons (p, apply_sub (cdr (l), s, on));
}

static list
apply_result (list l) {
  if (nil (l)) return l;
  proof p= as_proof (car (l));
  return cons (p->f, apply_result (cdr (l)));
}

proof
converter_rep::apply (generic f, proof from_p, int mode) {
  if (is_ambiguous (from_p)) {
    list l= from_p->deps, sols;
    while (!nil (l)) {
      proof p= apply (f, as_proof (car (l)), mode);
      if (nil (p));
      else if (is_ambiguous (p)) sols= p->deps * sols;
      else sols= cons (p, sols);
      l= cdr (l);
    }
    return ambiguous_proof (cons (ID_AND, apply_result (sols)), sols);
  }

#ifdef AMBIGUOUS_FIX
  // FIXME: all this is really dirty...
  // This fix is needed for assignments v[i] := x (v: Vector) to be OK.
  // It treats simple function application in a more direct way
  if (is_arrow (f)) {
    generic from= from_p->f;
    proof to_p= convert (from, f[1], mode);
    if (!nil (to_p) && !is_ambiguous (to_p))
      return proof (f[2], symbol ("apply"), list (from_p, to_p), to_p->inst);
      // FIXME: what to do with from_p->inst ?
  }
#endif

  generic from= from_p->f;
  generic fv  = free_var (symbol ("im"));
  generic g   = arrow (from, fv);
  proof p= convert (f, g, mode);
  if (nil (p)) return proof ();
  list l= instantiate (p, fv);
  l= apply_sub (l, "apply", from_p);
  l= requantify (l);
  return ambiguous_proof (cons (ID_AND, apply_result (l)), l);
}

proof
converter_rep::pullback (generic f, proof to_p, int mode) {
  if (is_ambiguous (to_p)) {
    list l= to_p->deps, sols;
    while (!nil (l)) {
      proof p= pullback (f, as_proof (car (l)), mode);
      if (nil (p));
      else if (is_ambiguous (p)) sols= p->deps * sols;
      else sols= cons (p, sols);
      l= cdr (l);
    }
    return ambiguous_proof (cons (ID_AND, apply_result (sols)), sols);
  }

  generic to= to_p->f;
  generic fv= free_var (symbol ("dom"));
  generic g = arrow (fv, to);
  proof p= convert (f, g, mode);
  if (nil (p)) return proof ();
  list l= instantiate (p, fv);
  l= apply_sub (l, "pullback", to_p);
  l= requantify (l);
  return ambiguous_proof (cons (ID_AND, apply_result (l)), l);
}

/******************************************************************************
* Implicit conversion
******************************************************************************/

static list
mapping (proof from, proof to) {
  list from_l, to_l;
  if (is_ambiguous (from)) from_l= from->deps;
  else from_l= list (from);
  if (is_ambiguous (to)) to_l= to->deps;
  else to_l= list (to);
  list l1= from_l, res;
  while (!nil (l1)) {
    list l2= to_l;
    while (!nil (l2)) {
      proof p1= as_proof (car (l1));
      proof p2= as_proof (car (l2));
      proof p (arrow (p1->f, p2->f), symbol ("map"), list (p1, p2), table ());
      res= cons (p, res);
      l2 = cdr (l2);
    }
    l1= cdr (l1);
  }
  return res;
}

proof
converter_rep::convert_implicit (generic from, generic to, int mode) {
  if ((mode & CV_DIRECT) != 0) return proof ();
  int i, j;
  array la, ra;
  la << generic (proof (from, symbol ("from"), list (), table ()));
  ra << generic (proof (to, symbol ("to"), list (), table ()));
  for (i=0; true; i++) {
    if (nil (la[i]) && nil (ra[i])) break;
    for (j=0; j<=i; j++)
      if ((!nil (la[i-j])) && (!nil (ra[j]))) {
	list sols;
	list l= mapping (as_proof (la[i-j]), as_proof (ra[j]));
	while (!nil (l)) {
	  proof q= as_proof (car (l));
	  proof p= convert (dcv, q->f, CV_DIRECT);
	  if (!nil (p)) {
	    proof r (from, to, "implicit", list (q, p));
	    if (is_ambiguous (r)) sols= r->deps * sols;
	    else sols= cons (r, sols);
	  }
	  l= cdr (l);
	}
	if (!nil (sols)) return ambiguous_proof (implies (from, to), sols);
      }
    if (nil (la[i])) la << proof ();
    else la << apply (lcv, as_proof(la[i]), CV_DIRECT);
    if (nil (ra[i])) ra << proof ();
    else ra << pullback (rcv, as_proof(ra[i]), CV_DIRECT);
  }
  return proof ();
}

/******************************************************************************
* Main conversion routines
******************************************************************************/

proof
converter_rep::convert (generic from, generic to, int mode) {
  // cout << "Converting " << from << " => " << to << lf;
  if (to == from) return proof (from, to, table ());

  /***/
  if (is_or (from)) return convert_from_or (from, to, mode);
  if (is_and (to)) return convert_to_and (from, to, mode);
  if (is_exists (from)) return convert_from_exists (from, to, mode);
  if (is_forall (to)) return convert_to_forall (from, to, mode);
  if (is_forall (from)) return convert_from_forall (from, to, mode);
  if (is_exists (to)) return convert_to_exists (from, to, mode);
  if (is_and (from)) return convert_from_and (from, to, mode);
  if (is_or (to)) return convert_to_or (from, to, mode);
  table t;
  if (match (from, to, t)) return proof (from, to, t);

  /***/
  if (is_arrow (from) && is_arrow (to))
    return convert_function (from, to, mode);
  if (is_across (from) && is_across (to) &&
      (from[0] == to[0]) &&
      (N(from) == N(to)))
    return convert_cross (from, to, mode);
  if ((is_cross (from) && is_tupletype (to)) ||
      (is_vcross (from) && is_vtupletype (to)))
    return convert_tuple (from, to, mode);
  if (is_aliastype (from))
    return convert_from_alias (from, to, mode);
  if (is_cross (from) && is_aliastype (to) &&
      is_cross (to[1]) && (N(from)==N(to[1])))
    return convert_aliases (from, to, mode);
  if (is_asinglet (from))
    return convert_from_singlet (from, to, mode);
  if (is_asinglet (to) || is_atupletype (to))
    return convert_to_singlet (from, to, mode);

  //This rule was a bit too zealous;
  //we now explicitly ignore void return values in conv_apply.cc
  //if (is_across (to) && (N(to) == 1))
  //return convert_ignore (from, to, mode);

  /***/
  return convert_implicit (from, to, mode);
}

generic
convert_prepare (generic g) {
  if (nil (g)) return g;
  if (is_list (g))
    return cons (convert_prepare (car (g)), convert_prepare (cdr (g)));
  if (g == MMX_CROSS) return ID_CROSS;
  if (g == MMX_VCROSS) return ID_VCROSS;
  if (g == MMX_TUPLETYPE) return ID_TUPLETYPE;
  if (g == MMX_VTUPLETYPE) return ID_VTUPLETYPE;
  if (g == MMX_FUNCTION) return ID_FUNCTION;
  if (g == MMX_ALIAS) return ID_ALIAS;
  if (g == MMX_FORALL) return ID_FORALL;
  if (g == MMX_EXISTS) return ID_EXISTS;
  if (g == MMX_IMPLIES) return ID_IMPLIES;
  if (g == MMX_AND) return ID_AND;
  if (g == MMX_OR) return ID_OR;
  if (g == MMX_NOT) return ID_NOT;
  if (g == MMX_VOID) return ID_VOID;
  return g;
}

void
converter_rep::test (generic f) {
  if (is_func (f, MMX_BEGIN, 1)) {
    int i, n= N(f);
    for (i=1; i<n; i++)
      test (f[i]);
    return;
  }
  f= convert_prepare (f);

  if (is_implies (f)) {
    cout << hrule;
    proof p= convert (f[1], f[2], CV_NORMAL);
    // cout << hrule;
    cout << p;
  }
  if (is_func (f, "cv", 1)) add (f[1], false, false);
  if (is_func (f, "lcv", 1)) add (f[1], true, false);
  if (is_func (f, "rcv", 1)) add (f[1], false, true);
  if (is_func (f, "lrcv", 1)) add (f[1], true, true);
  if (is_func (f, "apply", 2)) {
    cout << hrule;
    proof q (f[2], symbol ("given"), list (), table ());
    proof p= apply (f[1], q, CV_NORMAL);
    // cout << hrule;
    cout << p;
  }
  if (is_func (f, "back", 2)) {
    cout << hrule;
    proof q (f[2], symbol ("given"), list (), table ());
    proof p= pullback (f[1], q, CV_NORMAL);
    // cout << hrule;
    cout << p;
  }
  if (is_func (f, "prg", 2)) {
    cout << hrule;
    cout << apply_prog (f[1], f[2]) << lf;
  }
  cout << hrule;
}

} // namespace mmx
