
/******************************************************************************
* MODULE     : simplify.cc
* DESCRIPTION: Rudimentary simplification of expression used before printing
*              Not much should be done so that the original intention of
*              the way things should be output is kept
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <simplify.hh>
#include <list.hh>
#include <symbols.hh>
#include <literal.hh>
namespace mmx {

GC_GLOBAL_STATIC (symbol, Zero, "0");
GC_GLOBAL_STATIC (symbol, One, "1");
GC_GLOBAL_STATIC (symbol, Minus_one, "-1");

static generic plus (generic g1, generic g2);
static generic minus (generic g1, generic g2);
generic simplify (generic g);

/******************************************************************************
* Additive arithmetic with simplifications
******************************************************************************/

static void
signed_decompose (generic g, generic& abs_g, int& sgn_g) {
  if (!is_list (g)) {
    if (is_symbol (g)) {
      signed_decompose (S (as_symbol (g)), abs_g, sgn_g);
      abs_g= symbol (abs_g);
    }
    else if (is_string (g)) {
      string s= as_string (g);
      if (N(s) > 1 && s[0] == '-' &&
	  (s[1] == '.' || (s[1] >= '0' && s[1] <= '9'))) {
	abs_g= generic (s (1, N(s)));
	sgn_g= -1;
      }
      else {
	abs_g= g;
	sgn_g= 1;
      }
    }
    else {
      abs_g= g;
      sgn_g= 1;
    }
  }
  else if (is_func (g, MMX_MINUS, 1)) {
    signed_decompose (g[1], abs_g, sgn_g);
    sgn_g= -sgn_g;
  }
  else if (is_func (g, MMX_TIMES, 2) || is_func (g, MMX_OVER, 2)) {
    generic a1, a2;
    int     s1, s2;
    signed_decompose (g[1], a1, s1);
    signed_decompose (g[2], a2, s2);
    abs_g= list (g[0], a1, a2);
    sgn_g= s1 * s2;
  }
  else {
    abs_g= g;
    sgn_g= 1;
  }
}

static generic
negate (generic g) {
  if (g == Zero) return g;
  if (is_func (g, MMX_MINUS, 1))
    return g[1];
  if (is_func (g, MMX_PLUS, 2))
    return minus (negate (g[1]), g[2]);
  if (is_func (g, MMX_MINUS, 2))
    return plus (negate (g[1]), g[2]);
  generic abs_g; int sgn_g;
  signed_decompose (g, abs_g, sgn_g);
  if (sgn_g > 0) return list (MMX_MINUS, abs_g);
  else return abs_g;
}

static generic
plus (generic g1, generic g2) {
  if (g1 == Zero) return g2;
  if (g2 == Zero) return g1;
  if (is_func (g2, MMX_MINUS, 1))
    return minus (g1, g2[1]);
  if (is_func (g2, MMX_PLUS, 2))
    return plus (plus (g1, g2[1]), g2[2]);
  if (is_func (g2, MMX_MINUS, 2))
    return minus (plus (g1, g2[1]), g2[2]);
  generic abs_g2; int sgn_g2;
  signed_decompose (g2, abs_g2, sgn_g2);
  if (sgn_g2 < 0) return list (MMX_MINUS, g1, abs_g2);
  else return list (MMX_PLUS, g1, abs_g2);
}

static generic
minus (generic g1, generic g2) {
  if (g2 == Zero) return g1;
  if (is_func (g2, MMX_MINUS, 1))
    return plus (g1, g2[1]);
  if (is_func (g2, MMX_PLUS, 2))
    return minus (minus (g1, g2[1]), g2[2]);
  if (is_func (g2, MMX_MINUS, 2))
    return plus (minus (g1, g2[1]), g2[2]);
  generic abs_g2; int sgn_g2;
  signed_decompose (g2, abs_g2, sgn_g2);
  if (g1 == Zero) {
    if (sgn_g2 > 0) return list (MMX_MINUS, abs_g2);
    else return abs_g2;
  }
  else {
    if (sgn_g2 > 0) return list (MMX_MINUS, g1, abs_g2);
    else return list (MMX_PLUS, g1, abs_g2);
  }
}

/******************************************************************************
* Multiplicative arithmetic with simplifications
******************************************************************************/

static generic
times (generic g1, generic g2) {
  if (g1 == Zero) return Zero;
  if (g2 == Zero) return Zero;
  if (g1 == One) return g2;
  if (g2 == One) return g1;
  if (g1 == Minus_one) return negate (g2);
  if (g2 == Minus_one) return negate (g1);
  return list (MMX_TIMES, g1, g2);
}

static generic
over (generic g1, generic g2) {
  if (g1 == Zero) return Zero;
  if (g2 == One) return g1;
  if (g2 == Minus_one) return negate (g1);
  return list (MMX_OVER, g1, g2);
}

static generic
power (generic g1, generic g2) {
  if (g1 == Zero) return Zero;
  if (g1 == One) return One;
  if (g2 == Zero) return One;
  if (g2 == One) return g1;
  return list (MMX_POWER, g1, g2);
}

/******************************************************************************
* Main simplification routine
******************************************************************************/

static list
simplify (list l) {
  if (nil (l)) return l;
  return cons (simplify (car (l)), simplify (cdr (l)));
}

generic
simplify (generic g) {
  if (!is_list (g)) {
    generic abs_g; int sgn_g;
    signed_decompose (g, abs_g, sgn_g);
    if (sgn_g > 0) return abs_g;
    else return list (MMX_MINUS, abs_g);
  }
  if (is_func (g, MMX_MINUS, 1))
    return negate (simplify (g[1]));
  if (is_func (g, MMX_PLUS, 2))
    return plus (simplify (g[1]), simplify (g[2]));
  if (is_func (g, MMX_MINUS, 2))
    return minus (simplify (g[1]), simplify (g[2]));
  if (is_func (g, MMX_TIMES, 2))
    return times (simplify (g[1]), simplify (g[2]));
  if (is_func (g, MMX_OVER, 2))
    return over (simplify (g[1]), simplify (g[2]));
  if (is_func (g, MMX_POWER, 2))
    return power (simplify (g[1]), simplify (g[2]));
  return simplify (as_list (g));
}

} // namespace mmx
