
/******************************************************************************
* MODULE     : gc.hh
* DESCRIPTION: Memory management and garbage collection
* COPYRIGHT  : (C) 2005  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_GC_HH
#define __MMX_GC_HH
#include <config.hh>
#include <stdlib.h>
#include <iostream>
#include <cassert>
using std::size_t;
using std::istream;
using std::ostream;
using std::cin;
using std::cout;
using std::cerr;

namespace mmx {

extern bool interrupt_evaluation;
extern bool gc_debug_flag;

class generic;
class generic_rep;
class string;
class list;
class gentable;
class table;
class symbol;
class mirror;
class value;
class keyword;
class converter;
class evaluator;

typedef void* void_ptr;
typedef char* char_ptr;
typedef char char_array[];
typedef generic (*mmx_function) (generic);

/******************************************************************************
* Registration of root pointers for the GC
******************************************************************************/

#define SOURCE_TRACKING

void gc_register (char* name, void* ptr);
void gc_unregister (char* name, void* ptr);
void mmx_register (char* name, void* ptr);
void mmx_unregister (char* name, void* ptr);
void mmx_register_char_array (char* name, char* ptr);
void mmx_unregister_char_array (char* name, char* ptr);

template<typename T>
class global_entry {
  char* name;
  T* ptr;
public:
  inline global_entry (char* name2, T* ptr2):
    name (name2), ptr (ptr2) { assert (false); }
  inline ~global_entry () { assert (false); }
};

#define GC_REGISTER_TYPE(T)						\
class global_entry<T > {						\
  char* name;								\
  T* ptr;								\
public:									\
  inline global_entry<T > (char* name2, T* ptr2):			\
    name (name2), ptr (ptr2) { gc_register (name, (void*) ptr); }	\
  inline ~global_entry<T > () { gc_unregister (name, (void*) ptr); }	\
};

template<> GC_REGISTER_TYPE (generic);
template<> GC_REGISTER_TYPE (string);
template<> GC_REGISTER_TYPE (list);
template<> GC_REGISTER_TYPE (gentable);
template<> GC_REGISTER_TYPE (table);
template<> GC_REGISTER_TYPE (symbol);
template<> GC_REGISTER_TYPE (mirror);
template<> GC_REGISTER_TYPE (value);
template<> GC_REGISTER_TYPE (keyword);
template<> GC_REGISTER_TYPE (converter);
template<> GC_REGISTER_TYPE (evaluator);

#define MMX_REGISTER_TYPE(T)						\
class global_entry<T> {							\
  char* name;								\
  T* ptr;								\
public:									\
  inline global_entry<T> (char* name2, T* ptr2):			\
    name (name2), ptr (ptr2) { mmx_register (name, (void*) ptr); }	\
  inline ~global_entry<T> () { mmx_unregister (name, (void*) ptr); }	\
};

template<> MMX_REGISTER_TYPE (int);
template<> MMX_REGISTER_TYPE (bool);
template<> MMX_REGISTER_TYPE (long long int);
template<> MMX_REGISTER_TYPE (void*);
template<> MMX_REGISTER_TYPE (char*);

template<>
class global_entry<char_array> {
  char* name;
  char* ptr;
public:
  inline global_entry<char_array> (char* name2, char* ptr2):
    name (name2), ptr (ptr2) { mmx_register_char_array (name, ptr); }
  inline ~global_entry<char_array> () {
    mmx_unregister_char_array (name, ptr); }
};

#define GC_GLOBAL(type,name,init)				\
  type name= init;						\
  static mmx::global_entry<type > reg_##name (#name, &name);

#define GC_GLOBAL_STATIC(type,name,init)			\
  static type name= init;					\
  static mmx::global_entry<type > reg_##name (#name, &name);

#define GC_GLOBAL_STATIC_CHAR_ARRAY(name,init)			\
  static char_array name= init;					\
  static mmx::global_entry<char_array> reg_##name (#name, name);

#define GC_LOCAL(type,name,init)				\
  type name= init;						\
  mmx::global_entry<type > reg_##name (#name, &name);

/******************************************************************************
* Marking objects
******************************************************************************/

void gc_mark (generic_rep* rep);

#define NO_MARKING(type)			\
inline void mark (type& x) {			\
  (void) x;					\
}

#define GC_MARKING(type)						\
inline friend void mark (type& x) {					\
  mmx::generic_rep* rep= ((mmx::generic*) ((void*) (&x))) -> rep;	\
  if (rep != NULL) gc_mark (rep);					\
}

NO_MARKING(int);
NO_MARKING(long long int);
NO_MARKING(bool);
NO_MARKING(void_ptr);
NO_MARKING(char_ptr);
NO_MARKING(ostream);
NO_MARKING(mmx_function);

/******************************************************************************
* Memory allocation
******************************************************************************/

void* mmx_malloc (size_t sz);
void mmx_free (void* ptr, size_t sz);
int mmx_used_bytes ();

/*
void* var_mmx_malloc (size_t sz);
void var_mmx_free (void* ptr, size_t sz);
} // quit namespace mmx
inline void* operator new (size_t sz) {
  return mmx::var_mmx_malloc (sz); }
inline void* operator new[] (size_t sz) {
  return mmx::var_mmx_malloc (sz); }
inline void operator delete (void* ptr, size_t sz) {
  mmx::var_mmx_free (ptr, sz); }
inline void operator delete[] (void* ptr, size_t sz) {
  mmx::var_mmx_free (ptr, sz); }
namespace mmx { // re-enter namespace mmx
*/

#ifdef SUPPORTS_GC

void* gc_malloc (size_t sz);
void gc_free (void* ptr, size_t sz);
void gc_collect ();

void gc_verify_allocated (int n= 1000000);
void gc_verify_available (void* addr, size_t sz);
int  gc_count_objects (int kind);
void gc_show_objects (int kind);
void gc_inspect ();
void gc_memory_dump ();

#define MMX_ALLOCATORS						\
public:								\
  inline void* operator new (size_t sz) {			\
    return mmx_malloc (sz); }					\
  inline void* operator new[] (size_t sz) {			\
    return mmx_malloc (sz); }					\
  inline void operator delete (void* ptr, size_t sz) {		\
    mmx_free (ptr, sz); }					\
  inline void operator delete[] (void* ptr, size_t sz) {	\
    mmx_free (ptr, sz); }

#define GC_ALLOCATORS						\
public:								\
  inline void* operator new (size_t sz) {			\
    return gc_malloc (sz); }					\
  inline void* operator new[] (size_t sz) {			\
    return gc_malloc (sz); }					\
  inline void operator delete (void* ptr, size_t sz) {		\
    gc_free (ptr, sz); }					\
  inline void operator delete[] (void* ptr, size_t sz) {	\
    gc_free (ptr, sz); }

#else // not SUPPORTS_GC

#define MMX_ALLOCATORS				\
public:

#define GC_ALLOCATORS				\
public:

#endif // SUPPORTS_GC

} // namespace mmx

#endif // __MMX_GC_HH
