
/******************************************************************************
* MODULE     : generic.hh
* DESCRIPTION: Abstract types
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __GENERIC_HH
#define __GENERIC_HH
#include <gc.hh>

#define AMBIGUOUS_FIX

namespace mmx {

class generic;

/******************************************************************************
* Exhaustive list of supported types
******************************************************************************/

#define UNKNOWN_TYPE          0
#define NIL_TYPE              1
#define STRING_TYPE           2
#define SYMBOL_TYPE           3
#define MIRROR_TYPE           4
#define ALIAS_TYPE            5
#define ALIASES_TYPE          6
#define ARRAY_TYPE            7
#define LIST_TYPE             8
#define TABLE_TYPE            9
#define XTABLE_TYPE          10
#define FUNCTION_TYPE        11
#define ERROR_TYPE           12
#define SINGLE_TYPE          13
#define POINTER_TYPE         14
#define PRINTER_TYPE         15
#define SERIALIZER_TYPE      16
#define VALUE_TYPE           17
#define LITERAL_INT_TYPE     18
#define LITERAL_FLOAT_TYPE   19
#define LITERAL_STRING_TYPE  20

#define EVALUATOR_TYPE       21
#define KEYWORD_TYPE         22
#define QUANTOR_TYPE         23
#define PROOF_TYPE           24
#define CONVERTER_TYPE       25

#define CCPP_PRINTER_TYPE    26
#define PRETTY_PRINTER_TYPE  27
#define EXTENDER_TYPE        28

#define GLUED_TYPE           29
#define GLUED_TYPE_TYPE      30
#define TYPE_CONVERTER_TYPE  31
#define XTUPLE_TYPE          32
#define XALIAS_TYPE          33

/******************************************************************************
* The generic class
******************************************************************************/

class generic_rep {
GC_ALLOCATORS;
public:
  inline generic_rep () {}
  inline virtual ~generic_rep () {}

  void index (int begin, int end);
  void index (generic g);
  int  begin ();
  int  end   ();

  virtual size_t Size () = 0;
  virtual void Mark () = 0;
  virtual int Type ();
  virtual int Hash ();
  virtual bool Equal (generic g);
  virtual generic Copy ();
  virtual int Arity ();
  virtual generic& Access (int i);
  virtual generic Range (int begin, int end);
  virtual generic& Car ();
  virtual generic Cdr ();
  virtual void Print (ostream& out);

  virtual bool Is_genalias ();
  virtual bool Is_gentable ();
  virtual bool Is_literal ();
  virtual void Assign_me (generic g);
};

class generic {
MMX_ALLOCATORS;
GC_MARKING(generic);
public:
  generic_rep* rep;

  inline generic (): rep (NULL) {}
  inline generic (generic_rep* rep2): rep (rep2) {}
  inline generic_rep* operator -> () { return rep; }
  inline bool operator == (generic g) {
    if (rep == NULL) return g.rep == NULL;
    return (rep == g.rep) || rep->Equal (g); }
  inline bool operator != (generic g) {
    if (rep == NULL) return g.rep != NULL;
    return (rep != g.rep) && (!rep->Equal (g)); }
  inline generic& operator [] (int i) { return rep->Access (i); }
  inline generic operator () (int beg, int end) {
    return rep->Range (beg, end); }
};

inline bool nil (generic g) { return g.rep == NULL; }
inline int gentype (generic g) {
  if (g.rep == NULL) return NIL_TYPE;
  return g.rep->Type (); }
inline int hash (generic g) {
  if (g.rep == NULL) return 0;
  return g.rep->Hash (); }
inline int N (generic g) {
  if (g.rep == NULL) return 0;
  return g.rep->Arity (); }
inline generic& car (generic g) { return g.rep->Car (); }
inline generic cdr (generic g) { return g.rep->Cdr (); }
inline generic copy (generic g) {
  if (g.rep == NULL) return g;
  return g->Copy (); }
inline ostream& operator << (ostream& out, generic g) {
  if (nil (g)) return out << "()";
  g.rep->Print (out);
  return out; }
generic& S (generic g);
generic extend (generic g, generic with); // really generic& g
void locate (generic g, int& begin, int& end);

/******************************************************************************
* Predicates
******************************************************************************/

inline bool is_alias (generic g) {
  return (!nil (g)) && (gentype (g) == ALIAS_TYPE); }
inline bool is_aliases (generic g) {
  return (!nil (g)) && (gentype (g) == ALIASES_TYPE); }
inline bool is_array (generic g) {
  return (!nil (g)) && (gentype (g) == ARRAY_TYPE); }
inline bool is_function (generic g) {
  return (!nil (g)) && (gentype (g) == FUNCTION_TYPE); }
inline bool is_genalias (generic g) {
  return (!nil (g)) && g->Is_genalias (); }
inline bool is_gentable (generic g) {
  return (!nil (g)) && g->Is_gentable (); }
inline bool is_list (generic g) {
  return nil(g) || (gentype(g) == LIST_TYPE); }
inline bool is_literal (generic g) {
  return (!nil (g)) && g->Is_literal (); }
inline bool is_mirror (generic g) {
  return (!nil (g)) && (gentype (g) == MIRROR_TYPE); }
inline bool is_pointer (generic g) {
  return (!nil (g)) && (gentype (g) == POINTER_TYPE); }
inline bool is_printer (generic g) {
  return (!nil (g)) && (gentype (g) == PRINTER_TYPE); }
inline bool is_serializer (generic g) {
  return (!nil (g)) && (gentype (g) == SERIALIZER_TYPE); }
inline bool is_single (generic g) {
  return (!nil (g)) && (gentype (g) == SINGLE_TYPE); }
inline bool is_string (generic g) {
  return (!nil (g)) && (gentype (g) == STRING_TYPE); }
inline bool is_symbol (generic g) {
  return (!nil (g)) && (gentype (g) == SYMBOL_TYPE); }
inline bool is_table (generic g) {
  return (!nil (g)) && (gentype (g) == TABLE_TYPE); }
inline bool is_value (generic g) {
  return (!nil (g)) && (gentype (g) == VALUE_TYPE); }
inline bool is_xtable (generic g) {
  return (!nil (g)) && (gentype (g) == XTABLE_TYPE); }

inline bool is_ccpp_printer (generic g) {
  return (!nil (g)) && (gentype (g) == CCPP_PRINTER_TYPE); }
inline bool is_pretty_printer (generic g) {
  return (!nil (g)) && (gentype (g) == PRETTY_PRINTER_TYPE); }
inline bool is_extender (generic g) {
  return (!nil (g)) && (gentype (g) == EXTENDER_TYPE); }
inline bool is_quantor (generic g) {
  return (!nil (g)) && (gentype (g) == QUANTOR_TYPE); }
inline bool is_proof (generic g) {
  return (!nil (g)) && (gentype (g) == PROOF_TYPE); }
inline bool is_converter (generic g) {
  return (!nil (g)) && (gentype (g) == CONVERTER_TYPE); }
inline bool is_keyword (generic g) {
  return (!nil (g)) && (gentype (g) == KEYWORD_TYPE); }
inline bool is_evaluator (generic g) {
  return (!nil (g)) && (gentype (g) == EVALUATOR_TYPE); }

inline bool is_glued (generic g) {
  return (!nil (g)) && (gentype (g) == GLUED_TYPE); }
inline bool is_glued_type (generic g) {
  return (!nil (g)) && (gentype (g) == GLUED_TYPE_TYPE); }
inline bool is_type_converter (generic g) {
  return (!nil (g)) && (gentype (g) == TYPE_CONVERTER_TYPE); }
inline bool is_xalias (generic g) {
  return (!nil (g)) && (gentype (g) == XALIAS_TYPE); }
inline bool is_xtuple (generic g) {
  return (!nil (g)) && (gentype (g) == XTUPLE_TYPE); }

inline bool is_func (generic g, generic s) {
  return is_list (g) && (N(g) > 0) && (car(g) == s); }
inline bool is_func (generic g, generic s, int n) {
  return is_list (g) && (N(g) == (n+1)) && (car(g) == s); }

/******************************************************************************
* Useful macro for defining abstract derived types of generic
******************************************************************************/

#define DEFAULT_METHODS(type)						\
MMX_ALLOCATORS;								\
GC_MARKING(type);							\
protected:								\
  type##_rep* rep;							\
public:									\
  inline type (type##_rep* rep2): rep (rep2) {}				\
  inline friend type as_##type (mmx::generic g) {			\
    if (!is_##type (g)) type_mismatch (g, #type);			\
    return type ((type##_rep*) g.rep); }				\
  inline operator mmx::generic () { return mmx::generic (rep); }	\
  inline type##_rep* operator -> () const { return rep; }		\
  friend class type##_rep;

#define DEFAULT_METHODS_PROTO(type)				\
MMX_ALLOCATORS;							\
GC_MARKING(type);						\
protected:							\
  type##_rep* rep;						\
public:								\
  inline type (type##_rep* rep2): rep (rep2) {}			\
  inline friend type as_##type (generic g) {			\
    if (!is_##type (g)) type_mismatch (g, #type);		\
    return type ((type##_rep*) g.rep); }			\
  inline operator generic ();					\
  inline type##_rep* operator -> () const { return rep; }	\
  friend class type##_rep;

#define DEFAULT_METHODS_CODE(type)				\
inline type::operator generic () { return generic (rep); }

#define DEFAULT_METHODS_PROTO_1(type,A)				\
MMX_ALLOCATORS;							\
GC_MARKING(type);						\
protected:							\
  type##_rep<A>* rep;						\
public:								\
  inline type (type##_rep<A>* rep2): rep (rep2) {}		\
  inline operator generic ();					\
  inline type##_rep<A>* operator -> () const { return rep; }	\
  friend class type##_rep<A>;

#define DEFAULT_METHODS_CODE_1(type,A)			\
template<typename A> type<A>::operator generic () {	\
  return generic (rep); }				\
template<typename A> type<A> as_##type (generic g) {	\
  if (!is_##type (g)) type_mismatch (g, #type);		\
  return type<A> ((type##_rep<A>*) g.rep); }

#define DEFAULT_METHODS_PROTO_2(type,A,B)			\
MMX_ALLOCATORS;							\
GC_MARKING(type);						\
protected:							\
  type##_rep<A,B>* rep;						\
public:								\
  inline type (type##_rep<A,B>* rep2): rep (rep2) {}		\
  inline operator generic ();					\
  inline type##_rep<A,B>* operator -> () const { return rep; }	\
  friend class type##_rep<A,B>;

#define DEFAULT_METHODS_CODE_2(type,A,B)				\
template<typename A, typename B> type<A,B>::operator generic () {	\
  return generic (rep); }						\
template<typename A, typename B> type<A,B> as_##type (generic g) {	\
  if (!is_##type (g)) type_mismatch (g, #type);				\
  return type<A,B> ((type##_rep<A,B>*) g.rep); }

/******************************************************************************
* Miscellaneous routines
******************************************************************************/

/* should be declared as volatile, but not recognized by all g++ compilers */
void fatal_error (char* message, char* routine, char* file);
void type_mismatch (generic g, char* expected_type);
void type_mismatch (generic g, char* type, char* rout, char* file);

enum print_format {
  blank, stroke, indent, unindent, lf, hrule,
  expand_cycles, compress_cycles,
  normal, condensed, detailed, indented
};

NO_MARKING(print_format);
template<> MMX_REGISTER_TYPE (print_format);

extern print_format current_print_format;
extern int current_indentation;
ostream& operator << (ostream& out, print_format x);
bool long_print_format ();
ostream& print (ostream& out, char* s, generic g1);
ostream& print (ostream& out, char* s, generic g1, generic g2);

inline int min (int i, int j) { return i<j? i: j; }
inline int max (int i, int j) { return i>j? i: j; }

}

#endif // __GENERIC_HH
